# 氦氪HEKR模块串口透传协议参考实现

## 简介

本项目为(氦氪HEKR模块串口透传协议)[http://docs.hekr.me/v4/%E7%A1%AC%E4%BB%B6%E5%BC%80%E5%8F%91/%E9%80%9A%E4%BF%A1%E5%8D%8F%E8%AE%AE/%E4%B8%B2%E5%8F%A3%E9%80%8F%E4%BC%A0%E5%8D%8F%E8%AE%AE/]的参考实现，用于演示使用48协议直接接入HEKR云的设备的具体消息交互流程。

## 软件需求

1. 主流发行版的GNU/Linux操作系统 或 Microsoft Windows 7/8/8.1/10 或 MacOS
2. GNU Make
3. GCC/Clang

### GNU/Linux

需要安装构建工具，例如基于`Ubuntu/Debian`的发行版本：

`apt-get install build-essential`

### Windows

需要安装(MinGW)[https://mingw-w64.org/]或类似的平台同时安装以上列出的构建工具。

### MacOS

安装Xcode Command Line Tools以及GNU Make。

## 源码结构

- `ra_types.h`：基本类型定义
- `hekr48trans.c`、 `hekr48trans.h`：与云端交互的消息构造接口。
- `md5sum.c`、`md5sum.h`：构建消息时使用的**MD5**实现

实际项目的开发过程中只需要用到以上几个文件对交互使用的消息进行构造或解析。

- `main.c`：演示程序入口
- `Makefile`：构建文件
- `README.md`：本说明文件
- `usocket.c`、`usocket.h`：演示程序使用的Socket封装。

以上文件只是给出一个使用消息构建接口的简单说明，实际项目中可以参考。


## 构建

解压缩源码包：

```
unzip hekr48passthrough_sample_c*.zip
cd hekr48passthrough_sample_c*
```

编译：

```
make
```

## demo运行

在HEKR云的(console)[https://console.hekr.me/]平台创建48透传协议的产品，批量生成设备后，填入一组*prodkey*、*devtid*、*devprikey*，观察消息收发过程。

```
./a.out <prodkey> <devtid> <devprikey>
```

## 使用说明

### 项目文件

在实际开发过程中，需要将以上提到的：

`ra_types.h`、`hekr48trans.c`、`hekr48trans.h`、`md5sum.c`、`md5sum.h`文件通过恰当的方式引入你的项目，可能是某个IDE的项目文件也可能是Makefile。

之后在项目的代码中包含`hekr48trans.h`头文件：

```
#include "hekr48trans.h"
```

即可使用该头文件提供的接口进行设备和云端交互所使用的消息的构造和解析。

### 接口概述

以下接口用于构造设备主动发给云端的消息：

```
h48_build_frame_verify_request
h48_build_frame_verify
h48_build_frame_heartbeat
h48_build_frame_devsend
h48_build_frame_appsendresp
```

以`h48_build_frame_verify_request`为例，该函数签名为：

```
int h48_build_frame_verify_request(ra_u8 *buf, const ra_size_t buf_size, const ra_u8 frame_num, \
        const char *prodkey, const char *devtid);
```

其中`buf`和`buf_size`分别为存放构造出的消息的缓冲区和缓冲区大小，`frame_num`为这条消息的帧序号，而`prodkey`和`devtid`分别为产品密钥和devtid。

传入所需参数并调用`h48_build_frame_verify_request`后就会在`buf`缓冲区中获得所需的帧的数据，并且该帧的长度作为该接口的返回值以`int`类型获得，此时调用设备平台上发送数据的接口（例如socket的send接口）发送数据即可。

### 通常程序流程

程序基本流程为**登陆**和**消息循环**两个步骤。

```
login(); // 登陆
while (true) { // 消息循环
    // 处理消息
}
```

其中登陆流程包括了使用从云端统一生成的几个值进行鉴权，消息循环则包括了处理从云端获取的消息以及根据需要主动上报消息。

### 登陆流程

```
// 登陆流程开始
data = h48_build_frame_verify_request(prodkey, devtid); // 构造登陆鉴权请求
send(data); // 发送登陆鉴权请求
data = recv() // 接收云端下发返回帧
randomkey = h48_extract_randomkey(data) // 从返回帧提取随机码
data = h48_build_frame_verify(randomkey, devtid, devprikey); // 构造鉴权
send(data); // 发送鉴权
data = recv(); // 接收鉴权返回
// 登陆流程完毕
```

### 消息循环流程

消息循环根据业务需求不同和具体平台不同而不同，一个典型的消息循环如下：

```
while (true) { // 消息循环
    if (need_heartbeat()) { // 如果需要心跳（默认心跳间隔为每30秒内需要一次心跳）
        data = h48_build_frame_heartbeat(); // 构造心跳帧
        send(data); // 发送心跳帧
    } 
    if (cloud_data_available()) { // 如果云端有消息下发
        data = recv(); // 接收消息
        data_type = h48_frame_decode_type(); // 解析消息类型
        if (data_type == H48_FRAME_TYPE_APPSEND) { // 如果消息类型为云端下发的控制帧
            data = h48_build_frame_appsendresp(); // 构造云端控制帧确认帧
            send(data); // 发送云端控制帧确认帧
        } else { // 如果消息类型为其它
            // 处理其它消息，消息类型包括如下：
            // H48_FRAME_TYPE_UNKNOWN 未知
            // H48_FRAME_TYPE_VERIFY_REQUEST_RESP 鉴权请求回复帧
            // H48_FRAME_TYPE_VERIFY_RESP 鉴权回复帧
            // H48_FRAME_TYPE_REPORTDEVINFO_RESP 设备信息上报回复帧
            // H48_FRAME_TYPE_APPSEND 云端下发控制帧
            // H48_FRAME_TYPE_DEVSEND_RESP 设备信息上报回复帧
            // H48_FRAME_TYPE_HEARTBEAT_RESP 心跳回复帧
            // 其中回复帧可以忽略
            process(data); // 解析和处理消息
        }
    }
    if (device_data_available()) { // 如果设备有消息上报
        data = h48_build_frame_devsend(dev_data); // 构造设备消息上报帧
        send(data); // 发送设备需要上报的消息
    }
}
```

### 分帧

因为和云端建立的TCP连接属于一种流式的连接，多个消息从云端下发时可能会导致多帧粘连，在消息发送密集程度不高的时候直接把每次收到的数据作为完整一帧进行解析通常情况下可以，但是在消息密集的时候可能需要手工分帧。

使用`h48_try_split_frame`接口对以收到的数据进行尝试解析，当解析出一个完整的帧则进行处理。

一个通常的从云端接收消息的模式类似如下：

```
while (true)
{
    data = recv_data_from_cloud(); // 从云端接收数据
    buffer = buffer + data; // 缓冲区追加新接收的数据
    if (h48_try_split_frame(buffer) > 0) // 如果数据包含了一个合法帧
    {
        process(buffer) // 处理该帧包含的命令
        pop_buffer(); // 从缓冲中去除已经处理的帧
    }
}
```

### 异常处理

因为网络不稳定或者下发的数据有误等原因可能会导致网络连接被断开，需要根据各自平台的特性自行处理异常（包括重发或重新登录等）。

**注意**：若以上描述有疑惑请参照示例代码`main.c`文件中各接口调用的流程。

## 示例程序操作

运行程序后进入登录流程，如果登录成功则会进入交互状态，在不输入任何数据至按回车的情况下会自动发送心跳，否则如果有数据则会发送数据上报帧。