#ifdef _MSC_VER
#define _CRT_SECURE_NO_WARNINGS
#endif

#include "platform.h" 
#if defined(PLATFORM_WINDOWS)
#include <Windows.h>
#ifdef _MSC_VER
#pragma comment(lib, "ws2_32.lib")
#endif
#else
#include <unistd.h>
#endif
#include <string.h>
#include <time.h>
#include <stdio.h>
#include "usocket.h"
#include "hekr48trans.h"

#define HUB_HOST "hub.hekr.me"
// #define HUB_HOST "test-hub.hekr.me"
#define HUB_PORT 87
#define FRAME48_BUF_LEN 255

typedef struct
{
    usocket_t *fd;
    ra_u16 msgid;
    ra_u8 frame_num;
    char *prodkey;
    char *devtid;
    char *devprikey;
} session;

static int appsendresp(session *sess, ra_u8 frame_num, ra_u16 msgid, const ra_u8 *apptid);

static int decode_received_data(session *sess, char *data, int len)
{
    int ret = 0;
    h48_frame_type type = h48_frame_decode_type((ra_u8 *)data, len);

    switch (type)
    {
        case H48_FRAME_TYPE_UNKNOWN:
            printf("unknown frame\n");
            break;
        case H48_FRAME_TYPE_VERIFY_REQUEST_RESP:
            printf("verify request resp\n");
            break;
        case H48_FRAME_TYPE_VERIFY_RESP:
            printf("verify resp\n");
            break;
        case H48_FRAME_TYPE_REPORTDEVINFO_RESP:
            printf("report devinfo resp\n");
            break;
        case H48_FRAME_TYPE_APPSEND:
            printf("appsend\n");
            /* Response */
            {
                ra_u8 frame_num;
                ra_u16 msgid;
                const ra_u8 *apptid = (const ra_u8 *)data + 5 * 2;

                hexlit_to_raw(&frame_num, (const char *)data + 3 * 2, 2);
                hexlit_to_raw16(&msgid, (const char *)data + 4 * 2, 4);

                appsendresp(sess, frame_num, msgid, apptid);
            }
            /* TODO: Do things with the data sent from cloud */
            break;
        case H48_FRAME_TYPE_DEVSEND_RESP:
            printf("devsend resp\n");
            break;
        case H48_FRAME_TYPE_HEARTBEAT_RESP:
            printf("heartbeat resp\n");
            break;
    }

    return ret;
}

static int login(session *sess)
{
    ra_u8 buf[FRAME48_BUF_LEN];
    int len;

    /* Send Verify Request */
    {
        if ((len = h48_build_frame_verify_request(buf, FRAME48_BUF_LEN, sess->frame_num++, sess->prodkey, sess->devtid)) < 0)
        {
            fprintf(stderr, "error: failed to build login frame\n");
            return -1;
        }

        printf("D->C: "); fwrite(buf, (size_t)len, 1, stdout); printf("\n");

        if (usocket_send(sess->fd, buf, (ra_size_t)len) < len)
        { fprintf(stderr, "error: failed to send verify request frame\n"); return -1; }

        if ((len = usocket_recv(sess->fd, buf, (ra_size_t)FRAME48_BUF_LEN)) < 0)
        { fprintf(stderr, "error: failed to receive verify request frame response\n"); return -1; }

        if (h48_frame_decode_type((ra_u8 *)buf, len) != H48_FRAME_TYPE_VERIFY_REQUEST_RESP)
        { fprintf(stderr, "error: invalid login frame response\n"); return -1; }

        printf("C->D: "); fwrite(buf, (size_t)len, 1, stdout); printf("\n");
        decode_received_data(sess, (char *)buf, len);
    }

    /* Verify */
    {
        char randomkey[32 + 1];

        /* Extract RandomKey */
        /* memcpy(randomkey, buf + 8, 32); */
        /* randomkey[32] = '\0'; */

        h48_extract_randomkey(randomkey, 32 + 1, buf, len);

        if ((len = h48_build_frame_verify(buf, FRAME48_BUF_LEN, sess->frame_num++, randomkey, sess->devtid, sess->devprikey)) < 0)
        { fprintf(stderr, "error: failed to build login frame\n"); return -1; }

        printf("D->C: "); fwrite(buf, (size_t)len, 1, stdout); printf("\n");

        if (usocket_send(sess->fd, buf, (ra_size_t)len) < len)
        { fprintf(stderr, "error: failed to send verify frame\n"); return -1; }

        if ((len = usocket_recv(sess->fd, buf, (ra_size_t)FRAME48_BUF_LEN)) < 0)
        { fprintf(stderr, "error: failed to receive verify frame response\n"); return -1; }

        if (h48_frame_decode_type((ra_u8 *)buf, len) != H48_FRAME_TYPE_VERIFY_RESP)
        { fprintf(stderr, "error: invalid verify response frame\n"); return -1; }

        printf("C->D: "); fwrite(buf, (size_t)len, 1, stdout); printf("\n");
        decode_received_data(sess, (char *)buf, len);
    }

    return 0;
}

static int heartbeat(session *sess)
{
    int len;
    ra_u8 buf[FRAME48_BUF_LEN];

    if ((len = h48_build_frame_heartbeat(buf, FRAME48_BUF_LEN, sess->frame_num++)) < 0)
    { fprintf(stderr, "error: failed to build heartbeat frame\n"); return -1; }

    printf("D->C: "); fwrite(buf, (size_t)len, 1, stdout); printf("\n");

    if (usocket_send(sess->fd, buf, (ra_size_t)len) < len)
    { fprintf(stderr, "error: failed to send heartbeat frame\n"); return -1; }

    if ((len = usocket_recv(sess->fd, buf, (ra_size_t)FRAME48_BUF_LEN)) <= 0)
    { fprintf(stderr, "error: failed to receive heartbeat frame response\n"); return -1; }

    printf("C->D: "); fwrite(buf, (size_t)len, 1, stdout); printf("\n");
    decode_received_data(sess, (char *)buf, len);

    return 0;
}

static int reportdevinfo(session *sess)
{
    int len;
    ra_u8 buf[FRAME48_BUF_LEN];
    static const char *data = "010000000000000000000000000000000000000000000000000000000000000000";

    if ((len = h48_build_frame_devsend(buf, FRAME48_BUF_LEN, sess->frame_num++, sess->msgid++, \
                    (const ra_u8 *)data, (ra_size_t)strlen(data))) < 0)
    { fprintf(stderr, "error: failed to build devsend frame\n"); return -1; }

    printf("D->C: "); fwrite(buf, (size_t)len, 1, stdout); printf("\n");

    if (usocket_send(sess->fd, buf, (ra_size_t)len) < len)
    { fprintf(stderr, "error: failed to send devsend frame\n"); return -1; }

    if ((len = usocket_recv(sess->fd, buf, (ra_size_t)FRAME48_BUF_LEN)) < 0)
    { fprintf(stderr, "error: failed to receive devsend frame response\n"); return -1; }

    printf("C->D: "); fwrite(buf, (size_t)len, 1, stdout); printf("\n");
    decode_received_data(sess, (char *)buf, len);

    return 0;
}

static int appsendresp(session *sess, ra_u8 frame_num, ra_u16 msgid, const ra_u8 *apptid)
{
    int len;
    ra_u8 buf[FRAME48_BUF_LEN];

    if ((len = h48_build_frame_appsendresp(buf, FRAME48_BUF_LEN, frame_num, msgid, apptid)) < 0)
    { fprintf(stderr, "error: failed to build appsendresp frame\n"); return -1; }

    printf("D->C: "); fwrite(buf, (size_t)len, 1, stdout); printf("\n");

    if (usocket_send(sess->fd, buf, (ra_size_t)len) < len)
    { fprintf(stderr, "error: failed to send appsendresp frame\n"); return -1; }

    return 0;
}

static int handle_appsend(session *sess)
{
    int ret = 0;
    ra_u8 buf[FRAME48_BUF_LEN];
    int len;

    ret = usocket_readable(sess->fd);
    if (ret == -1)
    { fprintf(stderr, "error: failed to test connection\n"); return -1; }
    else if (ret == 0) return 0;

    if ((len = usocket_recv(sess->fd, buf, (ra_size_t)FRAME48_BUF_LEN)) < 0)
    { fprintf(stderr, "error: failed to receive devsend frame response\n"); return -1; }

    printf("C->D: "); fwrite(buf, (size_t)len, 1, stdout); printf("\n");
    decode_received_data(sess, (char *)buf, len);

    return 0;
}

static int loop(session *sess)
{
    // int tick = 0;

    if (reportdevinfo(sess) != 0) { return -1; }

    if (heartbeat(sess) != 0) { return -1; }

    for (;;)
    {
        int outbound_data_len;
        char outbound_data[256];
        fgets(outbound_data, 256, stdin);
        outbound_data_len = strlen(outbound_data);
        if (outbound_data_len == 0)
        {
            if (heartbeat(sess) != 0) { return -1; }
        }
        else
        {
            if (outbound_data[outbound_data_len - 1] == '\n')
            {
                outbound_data[outbound_data_len - 1] = '\0';
                outbound_data_len--;
            }
            if (outbound_data_len == 0)
            {
                if (heartbeat(sess) != 0) { return -1; }
            }
            else
            {
                printf("D->C: "); fwrite(outbound_data, (size_t)outbound_data_len, 1, stdout); printf("\n");

                if (usocket_send(sess->fd, outbound_data, (size_t)outbound_data_len) < outbound_data_len)
                { fprintf(stderr, "error: failed to send devsend frame\n"); return -1; }
            }
        }

        /*
        tick += 1;
        if (tick == 20)
        {
            if (heartbeat(sess) != 0) { return -1; }
            tick = 0;
        }
        */

        handle_appsend(sess);

#if defined(PLATFORM_WINDOWS)
		Sleep(1000);
#else
		sleep(1);
#endif
    }

    return 0;
}

int main(int argc, char *argv[])
{
    usocket_t fd;
    session sess;

#if defined(PLATFORM_WINDOWS)
    {
        int ret;
        WSADATA wsaData;
        ret = WSAStartup(MAKEWORD(2, 2), &wsaData);
        if (ret != 0)
        { fprintf(stderr, "WSAStartup failed: %d\n", ret); return -1; }
    }
#endif

    if (argc != 4)
    {
        fprintf(stderr, "usage: %s <prodkey> <devtid> <devprikey>\n", argv[0]);
        return -1;
    }
    else
    {
        sess.prodkey = argv[1];
        sess.devtid = argv[2];
        sess.devprikey = argv[3];
    }

    /* Establish a TCP connection between your device and the 'hub' cloud */
    usocket_init(&fd);
    if (usocket_connect(&fd, HUB_HOST, HUB_PORT) != 0)
    {
        fprintf(stderr, "error: failed to connect to %s:%d\n", HUB_HOST, (int)HUB_PORT);
        return -1;
    }

    sess.fd = &fd;
    sess.frame_num = 0;
    sess.msgid = 0;

    /* Login */
    if (login(&sess) != 0) { goto fail; }

    /* Loop */
    loop(&sess);

    return 0;
fail:
    return -1;
}

